"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSServer = exports.TSLogType = void 0;
const http_1 = require("http");
const commons_1 = require("./commons");
const fs_1 = require("./fs");
const tsrequest_1 = require("./tsrequest");
const types_1 = require("./types");
const utils_1 = require("./utils");
var TSLogType;
(function (TSLogType) {
    TSLogType["Log"] = "Info";
    TSLogType["Warning"] = "Warning";
    TSLogType["Error"] = "Error";
})(TSLogType = exports.TSLogType || (exports.TSLogType = {}));
;
class TSServer {
    // =================== CONSTRUCTOR =======================
    constructor(endPoints, host, port, staticFolders = [], managedTypes = TSServer.standardTypes(), fileCacheSize = 256 * 1024, // default 256 ko cache file size max
    maxCacheSize = 128 * 1024 * 1024, // default 128 Mo of memory cache
    logger) {
        this._maxBlacklistedFiles = 50000; // after that amount of blacklisted static files we will slowing down to answer on not found files
        this._maxCachedFiles = 10000;
        this._logger = (server, _, type, message) => __awaiter(this, void 0, void 0, function* () {
            (0, utils_1.$logterm)(`&0&xfoundation-ts[&wminimal server&x:${server.port}]-` + TSServer.__logHeaders[type] + "&0 &w" + message + "&0");
        });
        staticFolders.forEach(p => {
            if (!(0, fs_1.$isdirectory)(p)) {
                throw `Static resource folder '${p}' was not found on disk.`;
            }
        });
        this.port = port;
        this.host = host;
        this._endPoints = endPoints;
        this._statics = staticFolders;
        this._managedTypes = managedTypes;
        this._cacheMemory = 0;
        this._cache = new Map();
        this._blacklisted = new Set();
        this._cachedFileSizeMax = fileCacheSize;
        this._maxCacheMemorySize = maxCacheSize;
        if ((0, commons_1.$ok)(logger)) {
            this._logger = logger;
        }
    }
    // =================== static methods =======================
    static start(endPoints, opts = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!TSServer.__server) {
                TSServer.__server = new TSServer(endPoints, (0, commons_1.$length)(opts.host) ? opts.host : 'http://localhost/', (0, commons_1.$isunsigned)(opts.port) && opts.port < types_1.UINT16_MAX ? opts.port : 3000, opts.staticFolders, opts.staticContentTypes, opts.maxCachedFileSize, opts.maxCacheSize);
                yield TSServer.__server._start();
            }
        });
    }
    static clearCaches() {
        return __awaiter(this, void 0, void 0, function* () {
            if (TSServer.__server) {
                yield TSServer.__server._clearCaches();
            }
        });
    }
    // returns undefined if the server is stoped.
    // after that you can do a new server start
    static stop() {
        return __awaiter(this, void 0, void 0, function* () {
            if (TSServer.__server) {
                const e = yield TSServer.__server._stop();
                if (!(0, commons_1.$ok)(e)) {
                    delete TSServer.__server;
                }
                return e;
            }
            return undefined;
        });
    }
    static standardTypes() {
        return {
            'html': 'text/html', 'htm': 'text/html',
            '.css': 'text/css',
            '.js': 'text/javascript',
            '.svg': 'image/svg+xml', '.svgz': 'image/svg+xml',
            'jpeg': 'image/jpeg', 'jpg': 'image/jpg',
            '.png': 'image/png',
            '.gif': 'image/gif',
            '.ico': 'image/x-icon',
            '.pdf': 'application/pdf',
        };
    }
    // =================== instance methods =======================
    _start() {
        return __awaiter(this, void 0, void 0, function* () {
            this._httpServer = (0, http_1.createServer)((req, res) => __awaiter(this, void 0, void 0, function* () {
                let status = undefined;
                try {
                    const url = new URL((0, commons_1.$string)(req.url), this.host); // new URL(request.url, `http://${request.headers.host}`);
                    const path = url.pathname;
                    if (!(0, commons_1.$length)(path)) {
                        status = tsrequest_1.Resp.Forbidden;
                        yield this._logger(this, req, TSLogType.Warning, 'root path not accessible.');
                        throw 'Root path is not Accessible';
                    }
                    // we manage static resources of certain kinds
                    const [staticResource, contentType] = this.getStaticResource(path);
                    if ((0, commons_1.$defined)(staticResource)) {
                        if (staticResource === null) {
                            const m = `impossible to read static resource '${path}'.`;
                            yield this._logger(this, req, TSLogType.Warning, m);
                            throw `Internal Error: ${m}`;
                        }
                        res.setHeader('Content-Type', contentType);
                        res.writeHead(tsrequest_1.Resp.OK);
                        res.end(staticResource);
                    }
                    const fn = this._endPoints[path];
                    if (!(0, commons_1.$ok)(fn)) {
                        status = tsrequest_1.Resp.NotFound;
                        const m = `resource '&r${path}&o' was not found.`;
                        yield this._logger(this, req, TSLogType.Warning, m);
                        throw m;
                    }
                    yield fn(url, req, res); // handling the response is user responsability
                    // every unhandled exception will be treated as Internal Error
                    yield this._logger(this, req, TSLogType.Log, `did handle resource '${path}'.`);
                }
                catch (e) {
                    let error = e.message;
                    if (!(0, commons_1.$length)(error)) {
                        error = 'Internal Error';
                    }
                    if (!(0, commons_1.$ok)(status)) {
                        status = tsrequest_1.Resp.InternalError;
                    }
                    res.setHeader('Content-Type', 'text/plain');
                    res.writeHead(status);
                    res.end(`${status} - ${error}`);
                }
            }));
            this._httpServer.listen(this.port, () => __awaiter(this, void 0, void 0, function* () {
                yield this._logger(this, undefined, TSLogType.Log, `running on port ${this.port} '${this.host}' ...`);
            }));
        });
    }
    getStaticResource(path) {
        if (this._statics.length) {
            let ret = undefined;
            const ext = (0, fs_1.$ext)(path);
            const type = (0, commons_1.$length)(ext) ? this._managedTypes[ext.toLowerCase()] : undefined;
            if ((0, commons_1.$defined)(type) && !this._blacklisted.has(path)) {
                ret = this._cache.get(path);
                if (!(0, commons_1.$defined)(ret)) {
                    for (let folder of this._statics) {
                        const file = (0, fs_1.$path)(folder, path);
                        if ((0, fs_1.$isfile)(file)) {
                            ret = (0, fs_1.$readBuffer)(file);
                            if ((0, commons_1.$ok)(ret) &&
                                ret.length <= this._cachedFileSizeMax &&
                                ret.length + this._cacheMemory <= this._maxCacheMemorySize &&
                                this._cache.size < this._maxCachedFiles) {
                                this._cache.set(path, ret);
                                this._cacheMemory += ret.length;
                            }
                            else if (!(0, commons_1.$ok)(ret) && this._blacklisted.size < this._maxBlacklistedFiles) {
                                // it will next time be considered as not found
                                this._blacklisted.add(path);
                            }
                            return [ret, type];
                        }
                    }
                    // forever marked as not found
                    if (this._blacklisted.size < this._maxBlacklistedFiles) {
                        this._blacklisted.add(path);
                    }
                }
                return [ret, type];
            }
        }
        return [undefined, 'text/plain'];
    }
    _clearCaches() {
        return __awaiter(this, void 0, void 0, function* () {
            this._cache.clear();
            this._blacklisted.clear();
        });
    }
    _stop() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._logger(this, undefined, TSLogType.Log, `server is exiting...`);
            if ((0, commons_1.$ok)(this._httpServer)) {
                const ret = yield _internalStopServer(this._httpServer);
                if ((0, commons_1.$ok)(ret)) {
                    yield this._logger(this, undefined, TSLogType.Error, `cannot stop for reason ${ret.name}:\n${ret.message}`);
                }
                return ret;
            }
            ;
            return undefined;
        });
    }
}
exports.TSServer = TSServer;
TSServer.__server = undefined;
TSServer.__logHeaders = {
    'Info': "&C&w LOGGING ",
    'Warning': "&O&w WARNING ",
    'Error': "&R&w  ERROR  "
};
const _internalStopServer = (server) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((resolve, reject) => {
        server.close((error) => {
            if ((0, commons_1.$ok)(error)) {
                return reject(error);
            }
            resolve(undefined);
        });
    });
});
//# sourceMappingURL=tsserver.js.map